module orderedList

  integer, parameter :: OL_SIZE = 24 ! Max amount of data storable in an oList

  type oList
     integer :: data(OL_SIZE)
     integer :: size = 0
  end type oList

  interface operator(+)
     module procedure union
     module procedure append
  end interface

  interface operator(-)
     module procedure diff_ol_ol
     module procedure diff_ol_i
  end interface

  interface assignment(=)
     module procedure assign
  end interface

  interface operator(==)
     module procedure equal
  end interface

  interface operator(.c.)
     module procedure contains_ol_i
  end interface
contains

  integer function olsize()
    ! Reutrn the defined oList size
    olsize = OL_SIZE
  end function olsize

  logical function contains_ol_i(s, i)
    type(oList), intent(in) :: s
    integer, intent(in)     :: i
    integer :: k
    contains_ol_i = .false.
    do k=1,s%size
       if (s%data(k) == i) then
          contains_ol_i = .true.
          return
       end if
    end do
  end function contains_ol_i

  integer function encode(ol,s)
    ! Convert an oList to a number
    type(oList), intent(in) :: ol
    integer, intent(in) :: s
    integer :: i
    encode = 0
    do i=ol%size,1,-1
       encode = encode + ol%data(i)*(s+1)**(ol%size-i)
    end do
!!$    print *,(ol%data(i),i=1,ol%size),' => ',encode
  end function encode
    
  type(oList) function decode(value,s)
    ! Convert a value to an oList, e.g. 38,9 -> 3,8
    integer, intent(in) :: value, s
    integer :: val, rem, i
    val = value
    i = 1
    do
       rem = mod(val,s+1)
       decode = decode + rem
       val = val/(s+1)
       i = i + 1
       if (val == 0) exit
    end do
!!$    print *,value,' <= ',(decode%data(i),i=1,decode%size)

  end function decode

  logical function equal(s1, s2)
    implicit none
    type(oList),  intent(in) :: s1
    type(oList),  intent(in) :: s2

    integer :: i
    equal = .true.
    do i=1,min(s1%size,s2%size)
       equal = equal .and. (s1%data(i) == s2%data(i))
    end do

  end function equal

  type(oList) function diff_ol_ol(s1, s2) result(diff)
    implicit none
    type(oList),  intent(in) :: s1
    type(oList),  intent(in) :: s2
    logical, dimension(:), allocatable :: allow
    integer :: i, j

    ! Initialize the allow array - .true. means element is allowed in diff
    allocate(allow(size(s1%data)))
    ! All elements in s1%data allowed initially
    allow = [ ((i <= s1%size),i=1,size(s1%data)) ]

    diff%size = 0
    do i=1,s1%size
       do j=1,s2%size
          if (s1%data(i) == s2%data(j)) then
             allow(i) = .false.
             exit
          end if
       end do
    end do
    diff%data = pack(s1%data,allow)
    diff%size = count(allow)
    deallocate(allow)
  end function diff_ol_ol

  type(oList) function diff_ol_i(s1, s2) result(diff)
    implicit none
    type(oList),  intent(in) :: s1
    integer,  intent(in) :: s2
    logical, dimension(:), allocatable :: allow
    integer :: i, j

    ! Initialize the allow array - .true. means element is allowed in diff
    allocate(allow(size(s1%data)))
    ! All elements in s1%data allowed initially
    allow = [ ((i <= s1%size),i=1,size(s1%data)) ]

    diff%size = 0
    do i=1,s1%size
       if (s1%data(i) == s2) then
          allow(i) = .false.
          exit
       end if
    end do
    diff%data = pack(s1%data,allow)
    diff%size = count(allow)
    deallocate(allow)
  end function diff_ol_i

  subroutine assign(s1, s2)
    ! s1 = s2
    type(oList),  intent(out) :: s1
    type(oList),  intent(in) :: s2
    
    s1%data = s2%data
    s1%size = s2%size
  end subroutine assign

  type(oList) function append(s1, d)
    implicit none
    ! s1 U s2
    type(oList), intent(in) :: s1
    integer, intent(in)   :: d
    integer :: i, j, k
    append = s1
!!$    write(70,*) 'adding ',d,' to ',(append%data(k),k=1,append%size)
    select case(append%size)
    case (0)
       append%size = append%size + 1
       append%data(append%size) = d
       return

    case(1)
       if (d < append%data(1)) then
          append%data(2) = append%data(1)
          append%data(1) = d
          append%size = append%size + 1
       elseif (d == append%data(1)) then
          return
       else
          append%size = append%size + 1
          append%data(append%size) = d
       end if

    case default
       if (d < append%data(1)) then
          k = 1

       elseif (d == append%data(1)) then
          return

       else
          k = append%size+1
          shuffle: do i=2,append%size
             if (d == append%data(i)) then
                return
             end if
             if ((append%data(i-1) < d) .and. (d < append%data(i))) then
                k = i
                exit shuffle
             end if
          end do shuffle
       end if

       ! Shuffle up to make room
       do j=append%size,k,-1
          append%data(j+1) = append%data(j)
       end do
       append%data(k) = d
       append%size = append%size + 1
             
    end select
  end function append

  type(oList) function union(s1, s2)
    implicit none
    ! s1 U s2
    !
    ! Create the union by interleaving
    type(oList),  intent(in) :: s1
    type(oList),  intent(in) :: s2

    logical   :: more_s1, more_s2
    integer   :: i, j, m

    if (s1%size == 0) then
       union = s2
       return
    elseif (s2%size == 0) then
       union = s1
       return
    end if
    union%size = 1
    if (s1%data(1) <= s2%data(1)) then
       union%data(1) = s1%data(1)
       i = 2
       j = 1
    else
       union%data(1) = s2%data(1)
       i = 1
       j = 2
    end if
    m = 2
    more_s1 = (i <= s1%size)
    more_s2 = (j <= s2%size)
    all: do
       if (more_s1 .and. more_s2) then
          if (s1%data(i) <= s2%data(j)) then
             if (s1%data(i) > union%data(m-1)) then
                union%data(m) = s1%data(i)
                m = m + 1
             end if
             i = i + 1
          else
             if (s2%data(j) > union%data(m-1)) then
                union%data(m) = s2%data(j)
                m = m + 1
             end if
             j = j + 1
          end if

       else if (more_s1 .and. s1%data(i) > union%data(m-1)) then
          union%data(m) = s1%data(i)
          i = i + 1
          m = m + 1

       else if (more_s2 .and. s2%data(j) > union%data(m-1)) then
          union%data(m) = s2%data(j)
          j = j + 1
          m = m + 1

       else
          i = i + 1
          j = j + 1
       end if
       more_s1 = (i <= s1%size)
       more_s2 = (j <= s2%size)
       if (.not. more_s1 .and. .not. more_s2) exit
    end do all

    union%size = m-1
  end function union

  subroutine test
    type(oList) :: s1, s2, s3
    integer     :: i
    ! Test diff
    s1%data = [ (i,i=1,OL_SIZE) ]
    s1%size = 3
    s2%data = [ (mod(2*i,OL_SIZE)+1,i=1,OL_SIZE) ]
    s2%size = 4
    s3 = s1 - s2
    write(*,1000) '     s1 = (',s1%size,') ',(s1%data(i),i=1,s1%size)
    write(*,1000) '     s2 = (',s2%size,') ',(s2%data(i),i=1,s2%size)
    write(*,1000) 's1 - s2 = (',s3%size,') ',(s3%data(i),i=1,s3%size)
    s3 = s1 - s1
    write(*,1000) 's1 - s1 = (',s3%size,') ',(s3%data(i),i=1,s3%size)
    s3 = s1 - 5
    write(*,1000) 's1 - 5  = (',s3%size,') ',(s3%data(i),i=1,s3%size)
    s3 = s1 - 2
    write(*,1000) 's1 - 2  = (',s3%size,') ',(s3%data(i),i=1,s3%size)

    ! Test union
    s3 = s1 + s2
    write(*,1000) 's1 + s2 = (',s3%size,') ',(s3%data(i),i=1,s3%size)
    s3 = s1 + s1
    write(*,1000) 's1 + s1 = (',s3%size,') ',(s3%data(i),i=1,s3%size)

1000 format(a,i0,a,9i3)

  end subroutine test
end module orderedList
